from .logging import logger, timelog
from .rules_db import Connection, Rule
from .entity import Entity, Doc
from . import engines
from .pipeline import Pipeline, PipelineStep
from .scanner import Scanner


__all__ = (
    "Connection",
    "Doc",
    "Entity",
    "Pipeline",
    "PipelineStep",
    "Rule",
    "Scanner",
    "engines",
    "logger",
    "timelog",
)



from typing import Optional, Any, List

import hyperscan as hs

from entityscan import Rule, Entity


class MatchHandler:
    def __init__(self, data: bytes, encoding: str, entities: List[Entity]):
        self.data = data
        self.encoding = encoding
        self.entities = entities

    def __call__(
        self,
        rule_id: int,
        start: int,
        end: int,
        flags: int,
        context: Optional[Any],
    ):
        text = self.data[start:end].decode(self.encoding)
        entity = Entity.from_rule(rule_id, text, start, end)
        self.entities.append(entity)


class PatternEngine:
    def __init__(self, encoding: str = None):
        self.encoding = encoding or "UTF-8"
        self.hyperscan_db = hs.Database()
        self.patterns = []
        self.ids = []
        self.flags = []
        self.is_active = False

    def add_rule(self, rule: Rule):
        pattern: bytes = rf"\b{rule.pattern}\b".encode(self.encoding)
        self.patterns.append(pattern)
        self.ids.append(rule.id)
        flag = hs.HS_FLAG_SOM_LEFTMOST
        if not rule.case_sensitive:
            flag |= hs.HS_FLAG_CASELESS
        self.flags.append(flag)
        self.is_active = True

    def compile(self):
        if self.is_active:
            self.hyperscan_db.compile(
                expressions=self.patterns,
                ids=self.ids,
                elements=len(self.patterns),
                flags=self.flags,
            )

    def scan(self, text: str):
        entities = []
        if self.is_active:
            data: bytes = text.encode(self.encoding)
            handler = MatchHandler(data, self.encoding, entities)
            self.hyperscan_db.scan(data, match_event_handler=handler)
        return entities



from itertools import chain
from typing import Generator

from dawg import BytesDAWG

from entityscan import Rule, Entity


class LiteralDAWG:
    def __init__(self, case_sensitive: bool):
        self.case_sensitive = case_sensitive
        self.data = []
        self.dawg = None

    def normalize(self, term: str):
        return term.lower() if not self.case_sensitive else term

    def add_data(self, term: str, rule_id: int):
        norm = self.normalize(term)
        bytes_id = rule_id.to_bytes(8, "little", signed=False)
        self.data.append((norm, bytes_id))

    def compile(self):
        self.dawg = BytesDAWG(self.data)

    def is_prefix(self, term: str):
        norm = self.normalize(term)
        for _ in self.dawg.iterkeys(norm):
            return True
        return False

    def iter_entities(
        self,
        term: str,
        start: int = None,
        end: int = None,
    ) -> Generator[Entity, None, None]:
        norm = self.normalize(term)
        for bytes_id in self.dawg.get(norm) or []:
            bytes_id = bytes_id[:-1]  # DAWG appending '>' to ints?
            rule_id = int.from_bytes(bytes_id, "little", signed=False)
            start = start or 0
            end = end or len(term)
            yield Entity.from_rule(rule_id, term, start, end)

    def scan(self, text: str) -> Generator[Entity, None, None]:
        last = 0
        curr = 0
        prefixes = []
        for right in chain(text, "\0"):
            if not right.isalnum():
                next_prefixes = []
                for left in chain([last], prefixes):
                    term = text[left:curr]

                    yield from self.iter_entities(term, left, curr)

                    if self.is_prefix(term):
                        next_prefixes.append(left)

                prefixes = next_prefixes
                curr += 1
                last = curr

            else:
                curr += 1


class LiteralEngine:
    def __init__(self):
        self.dawgs = {
            True: LiteralDAWG(case_sensitive=True),
            False: LiteralDAWG(case_sensitive=False),
        }

    def add_rule(self, rule: Rule):
        dawg = self.dawgs.get(bool(rule.case_sensitive))

        dawg.add_data(rule.name, rule.id)
        for synonym in rule.synonyms or []:
            dawg.add_data(synonym, rule.id)

    def compile(self):
        for dawg in self.dawgs.values():
            dawg.compile()

    def scan(self, text: str) -> Generator[Entity, None, None]:
        for dawg in self.dawgs.values():
            yield from dawg.scan(text)



import re
from collections import defaultdict
from functools import reduce
from string import punctuation
from typing import Iterable, List, Optional, Dict

from entityscan import Rule, Entity

# https://stackoverflow.com/a/64303618
Trie = lambda: defaultdict(Trie)


class CompositeFinder:
    def __init__(self, engine: "CompositeEngine", text: str, entities: list):
        self.engine = engine
        self.text = text
        self.entities = entities
        self.composites = []
        self.curr_prefixes = []
        self.next_prefixes = []

    def find(self):
        for this in self.entities:
            if self.engine.is_prefix((this,)):
                self.next_prefixes.append((this,))

            for prefix in self.curr_prefixes:
                t_start, p_end = this.start, prefix[-1].end

                # span clean is without the punctuation
                span_exact = self.text[slice(p_end, t_start)]
                span_clean = span_exact.strip(punctuation)
                span_exact = span_exact.strip()
                span_clean = span_clean.strip()

                dist = len(span_clean)

                if dist <= self.engine.max_dist:
                    self.next_prefixes.append(prefix)

                    self.process_prefix(prefix, span_exact, this)
                    if span_exact and (span_clean != span_exact):
                        self.process_prefix(prefix, span_clean, this)

            self.curr_prefixes = self.next_prefixes
            self.next_prefixes = []

        for prefix in self.curr_prefixes:
            for trail, label in self.engine.get_trail_text(prefix).items():
                last_end = prefix[-1].end
                if trail == self.text[slice(last_end, last_end + len(trail))]:
                    self.push_composite(label, prefix + (trail,))

        return self.composites

    def process_prefix(self, prefix: tuple, span: str, this: dict):
        if span:
            prefix += (span,)

        prefix += (this,)

        if self.engine.is_prefix(prefix):
            self.next_prefixes.append(prefix)

        label = self.engine.get_end(prefix)
        if label:
            self.push_composite(label, prefix)

    def push_composite(self, label, prefix):
        entities = [item for item in prefix if isinstance(item, Entity)]
        start, end = entities[0].start, entities[-1].end

        if isinstance(prefix[-1], str):
            end += len(prefix[-1])

        text = self.text[start:end]
        meta = Entity.collapse_meta(entities)
        composite = Entity(
            text=text,
            label=label,
            start=start,
            end=end,
            name=text,
            meta=meta,
            entities=entities,
        )
        self.composites.append(composite)


class CompositeEngine:

    # trie helpers

    END = object()

    def __init__(self):
        self.trie = Trie()
        self.max_dist = 1

    def __contains__(self, item):
        return item in self.trie

    def __getitem__(self, part: str):
        return self.trie.get(part)

    def __setitem__(self, parts: tuple, label):
        parts = [p.label if isinstance(p, Entity) else p for p in parts]
        val = reduce(dict.__getitem__, parts, self.trie)
        val[self.END] = label

    def is_prefix(self, parts: tuple):
        val = self.reduce(parts)
        prefix_keys = val and (val.keys() - {self.END})
        return len(prefix_keys or []) > 0

    def get_end(self, parts: tuple) -> Optional[str]:
        val = self.reduce(parts)
        return val and val.get(self.END)

    def get_trail_text(self, parts: tuple) -> Dict[str, str]:
        val = self.reduce(parts)
        prefix_keys = val and (val.keys() - {self.END})
        trails = {}
        for key in prefix_keys:
            label = val.get(key, {}).get(self.END)
            if label:
                trails[key] = label
        return trails

    def reduce(self, parts: tuple):
        parts = [p.label if isinstance(p, Entity) else p for p in parts]

        trie = self.trie
        for part in parts:
            # need to check first, because getting creates
            if part in trie:
                trie = trie.get(part)

            # if invalid, return None
            else:
                return None

        return trie

    # add composite pattern

    def add_rule(self, rule: Rule):
        parts, this_max_dist = self.to_parts(
            rule.pattern, rule.ignore_punctuation
        )
        self[parts] = rule.label
        self.max_dist = max(self.max_dist, this_max_dist)

    re_labels = re.compile(r"(@[A-Z_]+)")
    re_punctuation = re.compile(r"")

    @classmethod
    def to_parts(cls, pattern: str, ignore_punctuation: bool) -> tuple:
        parts = []
        prev = None
        max_dist = 0
        pattern = pattern.strip()

        for match in cls.re_labels.finditer(pattern):
            # remove leading @
            label = match.group()[1:]

            # current
            curr = dict(start=match.start(), end=match.end(), label=label)

            # if not first
            if prev:
                span = pattern[prev["end"] : curr["start"]]
                max_dist = max(max_dist, len(span))
                if span and ignore_punctuation:
                    span = span.strip(punctuation)
                span = span.strip()
                if span:
                    parts.append(span)

            parts.append(curr["label"])
            prev = curr

        if prev:
            span = pattern[prev["end"] :]
            max_dist = max(max_dist, len(span))
            if span and ignore_punctuation:
                span = span.strip(punctuation)
            span = span.strip()
            if span:
                parts.append(span)

        return tuple(parts), max_dist

    # scan

    def process(self, text: str, entities: Iterable) -> List[Entity]:
        keep_going = True

        while keep_going:
            keep_going = False
            entities = Entity.sort_filter(entities)
            composites = CompositeFinder(self, text, entities).find()

            if composites:
                entities = self.filter_sub_entities(composites, entities)
                keep_going = True

        return entities

    @classmethod
    def filter_sub_entities(cls, composites, entities):
        next_entities = []
        composites = Entity.sort_filter(composites)
        seen_offsets = set()

        for composite in composites:
            off = set(range(composite.start, composite.end + 1))

            if seen_offsets.isdisjoint(off):
                seen_offsets.update(off)
                next_entities.append(composite)

        for entity in entities:
            if entity.start not in seen_offsets:
                next_entities.append(entity)

        return next_entities



import json
from collections import defaultdict
from dataclasses import dataclass
from functools import cached_property
from typing import List, Iterable, Optional

from entityscan import Rule


class MetaDict:
    def __init__(self, data: dict):
        self._data = data or {}

    def __getattr__(self, key):
        return self._data.get(key)


@dataclass
class Entity:
    text: str
    label: str
    start: int = 0
    end: int = 0
    name: Optional[str] = None
    synonyms: Optional[List[str]] = None
    entities: Optional[List["Entity"]] = None
    meta: Optional[dict] = None

    def __lt__(self, other: "Entity"):
        return self.sort_key < other.sort_key

    def __hash__(self):
        return hash((self.text, self.label, self.start, self.end))

    def iter_descendents(self):
        for entity in self.entities or []:
            yield entity
            yield from entity.iter_descendents()

    def dict(self, keep_null=False):
        data = dict(
            (k, v)
            for k, v in vars(self).items()
            if keep_null or (v is not None) and k[0] != "_"
        )
        if self.entities:
            data["entities"] = [
                e.dict(keep_null=keep_null) for e in self.entities
            ]
        return data

    def json(self, indent=4):
        return json.dumps(self.dict(), indent=indent)

    @cached_property
    def _m(self):
        return MetaDict(self.meta or {})

    @property
    def sort_key(self):
        return self.start, (-1 * self.end), (-1 * len(self.entities or []))

    @classmethod
    def from_rule(cls, rule_id: int, text: str, start: int, end: int):
        rule: Rule = Rule.select(lambda r: r.id == rule_id).first()
        groups = rule.get_groups(text)
        meta = {**(rule.meta or {}), **(groups or {})} or None

        return cls(
            label=rule.label,
            text=text,
            start=start,
            end=end,
            meta=meta,
            name=rule.name or text,
            synonyms=rule.synonyms,
        )

    @classmethod
    def sort_filter(cls, entities: Iterable["Entity"]) -> List["Entity"]:
        """www
        Sorting is based on Entity.sort_key (+start, -end, -sub entity count)

        Filtering follows the following logic:
            - Composites cause sub-entities to be removed
            - Entities of the same label that overlap goes to the longer
        """

        entities = sorted(entities)
        max_end = defaultdict(int)
        filtered = []

        for entity in entities:
            if max_end[entity.label] < entity.end:
                max_end[entity.label] = max(max_end[entity.label], entity.end)

                for child in entity.iter_descendents():
                    max_end[child.label] = max(max_end[child.label], child.end)

                filtered.append(entity)

        return filtered

    @classmethod
    def collapse_meta(cls, entities: List["Entity"]):
        meta = defaultdict(list)

        for entity in entities:
            prefix = entity.label.lower()
            meta[prefix].append(entity.name)
            for k, v in (entity.meta or {}).items():
                meta[f"{prefix}_{k}"].append(v)

        return dict((k, v[0] if len(v) == 1 else v) for k, v in meta.items())


@dataclass
class Doc:
    text: str
    entities: Optional[List["Entity"]] = None



import os
import re
from functools import lru_cache
from urllib.parse import urlparse

from pony import orm

db = orm.Database()


@lru_cache()
def compile_pattern(pattern: str, case_sensitive: bool):
    flags = 0 if case_sensitive else re.IGNORECASE
    return re.compile(pattern, flags=flags)


class Rule(db.Entity):
    id = orm.PrimaryKey(int, auto=True)
    label = orm.Required(str)
    meta = orm.Optional(orm.Json, nullable=True)

    # regex (expression or composite)
    pattern = orm.Optional(str, nullable=True)

    # literals
    name = orm.Optional(str, nullable=True)
    synonyms = orm.Optional(orm.StrArray, nullable=True)

    # flags
    is_composite = orm.Optional(bool, sql_default=False)
    case_sensitive = orm.Optional(bool, sql_default=False)
    ignore_punctuation = orm.Optional(bool, sql_default=False)
    has_groups = orm.Optional(bool, sql_default=False)

    @property
    def is_pattern(self):
        return not self.is_literal and not self.is_composite

    @property
    def is_literal(self):
        return self.name is not None

    def get_groups(self, text: str):
        if self.has_groups:
            pattern = compile_pattern(self.pattern, self.case_sensitive)
            match = pattern.match(text)
            return match.groupdict()


class Connection:
    def __init__(self, db_url: str = None, create_tables: bool = False):
        db_url = db_url or os.environ.get("ENTITYSCAN_DB_URL")
        db_bind_params = parse_url(db_url)

        table = db_bind_params.pop("_table_", False)
        if table:
            Rule._table_ = table

        db.bind(**db_bind_params)
        db.generate_mapping(create_tables=create_tables)

    @classmethod
    def close(cls):
        db.disconnect()
        db.provider = None

    @orm.db_session
    def get_count(self):
        return Rule.select().count()

    @orm.db_session
    def delete(self):
        return Rule.select().delete()


def parse_url(db_url: str):
    # generate arguments for database binding:
    # https://docs.ponyorm.org/firststeps.html#database-binding

    url = urlparse(db_url)
    scheme = "postgres" if url.scheme == "postgresql" else url.scheme
    assert scheme in {"sqlite", "postgres"}, f"Unsupported: {url.scheme}"

    kw = {"provider": scheme}
    path_parts = url.path.split("/")

    if scheme == "sqlite":
        if path_parts[1].startswith(":"):
            kw["filename"] = path_parts[1]
        else:
            kw["filename"] = "/".join(path_parts)

    # postgres://user:pass@host:port/database/schema.table
    if scheme == "postgres":
        kw["user"] = url.username or ""
        kw["password"] = url.password or ""
        kw["host"] = url.hostname or ""

        path_parts += ["", "", ""]
        kw["database"] = path_parts[1]

        schema = path_parts[2] or ""
        table = path_parts[3] or ""
        if schema and table:
            kw["_table_"] = (schema, table)

        assert kw["database"], f"No database name provided in URL {db_url}"

    return kw



import re
from os import path
from pathlib import Path

from setuptools import find_packages, setup


def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    version = Path(package, "__version__.py").read_text()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", version).group(1)


def get_long_description():
    this_directory = path.abspath(path.dirname(__file__))
    with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
        long_description = f.read()
    return long_description


install_requires = [
    "hyperscan==0.2.0",
    "pony==0.7.16",
    "DAWG==0.8.0",
]

setup(
    name="entityscan",
    python_requires=">=3.6",
    version=get_version("src/entityscan"),
    author="Ian Maurer",
    author_email="ian@genomoncology.com",
    packages=find_packages("src/"),
    package_dir={"": "src"},
    include_package_data=True,
    install_requires=install_requires,
    description="Find entities using Hyperscan",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/genomoncology/entityscan",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering",
    ],
)



from pathlib import Path
from logging import INFO
from typing import Iterable, List
from itertools import chain
import json

from pony import orm

from entityscan import (
    Connection,
    Pipeline,
    Rule,
    Doc,
    Entity,
    engines,
    timelog,
    logger,
)


class Scanner:

    def __init__(self, connection: Connection = None, encoding: str = None):
        self.connection = connection
        self.encoding = encoding or "UTF-8"
        self.patterns = engines.PatternEngine(encoding=self.encoding)
        self.composites = engines.CompositeEngine()
        self.literals = engines.LiteralEngine()

    @orm.db_session
    def load_rules(self, rules_jsonl: Path):
        assert rules_jsonl.suffix == ".jsonl", "Only supports JSONL files."
        for line in rules_jsonl.open("r"):
            line = line.strip()
            if line and line.startswith("{"):
                try:
                    kw = json.loads(line)
                    Rule(**kw)
                except json.JSONDecodeError:
                    logger.exception(f"JSON Fail: {line[:30]}...{line[-30:]}")
                    raise

    @orm.db_session
    @timelog("compile", level=INFO)
    def compile(self):
        for rule in Rule.select():
            if rule.is_pattern:
                self.patterns.add_rule(rule)
            elif rule.is_composite:
                self.composites.add_rule(rule)
            elif rule.is_literal:
                self.literals.add_rule(rule)
            else:
                raise ValueError(f"Invalid rule: {rule.id}")

        self.patterns.compile()
        self.literals.compile()

        return self

    @orm.db_session
    def scan_expressions(self, text: str) -> List[Entity]:
        p_entities = self.patterns.scan(text)
        l_entities = self.literals.scan(text)
        return Entity.sort_filter(chain(p_entities, l_entities))

    def scan(self, text: str, pipeline: str = None) -> List[Entity]:
        entities = self.scan_expressions(text=text)
        entities = self.composites.process(text, entities)
        entities = Entity.sort_filter(entities)

        if pipeline:
            entities = Pipeline.run(pipeline, text, entities)
            entities = Entity.sort_filter(entities)
        return entities

    def parse(self, text: str, pipeline: str = None) -> Doc:
        entities = self.scan(text=text, pipeline=pipeline)
        return Doc(text=text, entities=entities)

    def find(self, text: str, labels: Iterable[str] = None) -> List[Entity]:
        entities = self.scan(text)
        if labels:
            labels = set(labels)
            entities = [ent for ent in entities if ent.label in labels]
        return entities

    def find_one(self, text: str, labels: Iterable[str] = None):
        entities = self.find(text=text, labels=labels)
        return entities[0] if len(entities) == 1 else None



from typing import List, Dict, Iterable

from entityscan import Entity


class PipelineStep:

    each_labels = {}

    def __call__(self, text: str, entities: Iterable[Entity]):
        entities = self.process_all(text, entities)
        for entity in entities:
            if entity.label in self.each_labels:
                yield from self.process_each(text, entity)
            else:
                yield entity

    # noinspection PyMethodMayBeStatic
    def process_all(self, text: str, entities: Iterable[Entity]):
        yield from entities

    # noinspection PyMethodMayBeStatic
    def process_each(self, text: str, entity: Entity):
        yield entity


class Pipeline:

    __pipelines: Dict[str, List[PipelineStep]] = None

    @classmethod
    def get_pipeline_steps(cls, name: str) -> Iterable[PipelineStep]:
        if Pipeline.__pipelines is None:
            Pipeline.__pipelines = {}
            for subclass in cls.__subclasses__():
                kw = vars(subclass)
                name = kw.get("name")
                if name:
                    Pipeline.__pipelines[name] = [
                        val
                        for val in kw.values()
                        if isinstance(val, PipelineStep)
                    ]

        return Pipeline.__pipelines.get(name, [])

    @classmethod
    def run(cls, pipeline_name: str, text: str, entities: List[Entity]):
        steps = Pipeline.get_pipeline_steps(pipeline_name)
        for step in steps:
            entities = step(text, entities)
        return entities



from time import perf_counter
from contextlib import contextmanager
import logging.config


logger = logging.getLogger("entityscan")
logger.setLevel("INFO")


@contextmanager
def timelog(msg: str, level=logging.INFO):
    pc = perf_counter()
    logger.log(msg=f"Start: {msg}", level=level)
    yield
    logger.log(msg=f"Done : {msg} [{perf_counter() - pc:.2f}s]", level=level)
